//*************************************************************************
//* ---------------------
//* CDCClusterMaker Class
//* ---------------------
//* (Description)
//*	This set of classes makes FADC clusters.
//* (Requires)
//*	TObjArray:
//*	JSFSteer:
//*	JSFModules:
//*	CDCRawAData:
//* (Provides)
//*	CDCClusterMaker		: steering class
//*	CDCClusterBuf		: buffer to store clusters
//*	CDCCluster		: cluster data class
//*	CDCFADCCalibration	: FADC calibration
//* (Update Record)
//*	1999/12/04  K.Fujii	Original version.
//*************************************************************************

#include "CDCClusterMaker.h"

Int_t CDCCluster::fNpres  = 20; 
Int_t CDCCluster::fNposts = 20; 
Int_t CDCCluster::fSortBy =  0; // (0,1) = (time,charge)

//_____________________________________________________________________
// ---------------------
// FADC Calibration Data
// ---------------------

ClassImp(CDCFADCCalibration)

CDCFADCCalibration::CDCFADCCalibration(Double_t *ped, Double_t *gain, 
  		     		Double_t *tped, Double_t *tscale, Int_t last)
{
  	CalibDatum *datum = new CalibDatum(ped,gain,tped,tscale,last);
  	Add(datum);
  	fCurr = datum;
}

CDCFADCCalibration::CDCFADCCalibration(Char_t * calibfile) 
{
	cout << " Going to open a calib file " << calibfile << endl;
	FILE *fd = fopen(calibfile,"r");
	cout << " Opened " << calibfile << endl;
	if (fd == 0) {
		Error("CDCFADCCalibration","Failed to open %s.",calibfile);
		return;
        }
        Int_t last = 99999999;
        Double_t ped[4], gain[4], tped[4], tscale[4];
        while ((fscanf(fd,
             "%d %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
        	 &last, &ped[0], &ped[1], &ped[2], &ped[3],
        		&gain[0], &gain[1], &gain[2], &gain[3],
        		&tped[0], &tped[1], &tped[2], &tped[3],
        		&tscale[0], &tscale[1], &tscale[2], &tscale[3])) != EOF) {
	   CalibDatum *datum 
	   	= new CalibDatum(ped,gain,tped,tscale,last);
	   Add(datum);
	}
	fclose(fd);
}

CDCFADCCalibration::~CDCFADCCalibration() { Delete(); }
  
Double_t CDCFADCCalibration::GetPedestal    (Int_t n, Int_t a) 
  		{ return fCurr->GetPedestal(n,a);    }
Double_t CDCFADCCalibration::GetGain        (Int_t n, Int_t a)
  		{ return fCurr->GetGain(n,a);        }
Double_t CDCFADCCalibration::GetTimePedestal(Int_t n, Int_t a)
  		{ return fCurr->GetTimePedestal(n,a);}
Double_t CDCFADCCalibration::GetTimeScale   (Int_t n, Int_t a)
  		{ return fCurr->GetTimeScale(n,a);   }

void CDCFADCCalibration::Update()
{
  	Int_t run = gJSF->GetRunNumber();
  	TIter next(this);
  	CalibDatum *datum;
  	while ((datum = (CalibDatum *)next())) {
  		fCurr = datum;
  		if (run <= fCurr->GetLastRun()) break;
  	}	
}


//_____________________________________________________________________
// -------------
// Cluster Datum
// -------------

ClassImp(CDCCluster)

CDCCluster::CDCCluster(CDCFADCCalibration *calib, Bool_t varped) : TObjArray(1), 
	fDone(kFALSE), fUseVarPed(varped), 
	fStationNo(0), fChannelNo(0),
	fStartTimeSlot(0), fLastTimeSlot(0),
	fSum(0.), fPeak(0.), fPedestal(0.),
	fStartTime(0.), fWidth(0.),
	fNpeds(10), 
	fCalib(calib) {}

CDCCluster::CDCCluster(CDCCluster & cl) : TObjArray(cl),
	fDone(cl.fDone), fUseVarPed(cl.fUseVarPed),
	fStationNo(cl.fStationNo), fChannelNo(cl.fChannelNo),
	fStartTimeSlot(cl.fStartTimeSlot), fLastTimeSlot(cl.fLastTimeSlot),
	fSum(cl.fSum), fPeak(cl.fPeak), fPedestal(cl.fPedestal),
	fStartTime(cl.fStartTime), fWidth(cl.fWidth),
	fNpeds(cl.fNpeds), 
	fCalib(cl.fCalib) {}

CDCCluster::~CDCCluster() {}

void CDCCluster::SetNpeds(Int_t n)
{
  	if (fNpeds != n) {
  		fNpeds = n;
  		fDone  = kFALSE;
  	}
  	if (!fDone) Update();
}

Double_t CDCCluster::GetPedestal()
{
  	if (!fUseVarPed) {
   		return fCalib->GetPedestal(fStationNo,fChannelNo);
  	} else {
  	if (!fDone) Update(); 
  		return fPedestal;
  	}
}

void CDCCluster::UseVarPed()
{
	if (!fUseVarPed) {
		fUseVarPed = kTRUE;
		fDone = kFALSE;
	}
	if (!fDone) Update();
}

void CDCCluster::UseConstPed()
{
	if (fUseVarPed) {
		fUseVarPed = kFALSE;
		fDone = kFALSE;
	}
	if (!fDone) Update();
}


Bool_t CDCCluster::Append(CDCRawADCDatum *adc)
{
	if (GetEntries() == 0) {
		fStationNo     = adc->GetStationNo();
		fChannelNo     = adc->GetChannelNo();
		fStartTimeSlot = adc->GetTime();
		fLastTimeSlot  = fStartTimeSlot;
		Add(adc);
		return kTRUE;
	} else {
		if (adc->GetStationNo() == fStationNo &&
		    adc->GetChannelNo() == fChannelNo &&
		    adc->GetTime() == fLastTimeSlot+1) {
			fLastTimeSlot++;
			Add(adc);
			return kTRUE;
		} else {
			return kFALSE;
		}
	}
			
}

void CDCCluster::Update()
{
	if (fDone) return;
	fDone = kTRUE;
	Double_t ped    = fCalib->GetPedestal(fStationNo,fChannelNo);
	Double_t gain   = fCalib->GetGain(fStationNo,fChannelNo);
	Double_t tped   = fCalib->GetTimePedestal(fStationNo,fChannelNo);
	Double_t tscale = fCalib->GetTimeScale(fStationNo,fChannelNo);

	TIter next(this);
	Int_t nslots = GetEntries();
	if (nslots <= fNpres+fNposts) {
		fPedestal  = ped;
		fPeak      = 0.;
		fSum       = 0.;
		fStartTime = (fStartTimeSlot - tped)*tscale;
		fWidth     = (nslots - fNpres - fNposts)*tscale;
		cerr << "???? CDCCluster::Update() ???? " 
		     << "cluster too narrow: "
		     << "Nslots = " << nslots << endl;
		DebugPrint("Detailed");
		return;
	}
	
	fPedestal = 0;
	CDCRawADCDatum *adcp;
	Int_t islot;
	for (islot = 0; islot < fNpres; islot++) {
		adcp = (CDCRawADCDatum *)next();
		if (islot < fNpeds) fPedestal += adcp->GetADC();
	}
	fPedestal /= fNpeds;
	
	if (fUseVarPed) ped = fPedestal;
	
	fSum   = 0;
	fPeak  = -99999.;
#if 0
	for (; islot < nslots - fNposts; islot++) {
		adcp = (CDCRawADCDatum *)next();
		Double_t adc = ped - adcp->GetADC();
		fPeak  = TMath::Max(fPeak,adc);
		fSum  += adc;
	}
	fSum *= gain;
	
	fStartTime = (fStartTimeSlot - tped)*tscale;
	fWidth     = (nslots - fNpres - fNposts)*tscale;
#endif
#if 0
	for (; islot < nslots; islot++) {
		adcp = (CDCRawADCDatum *)next();
		Double_t adc = ped - adcp->GetADC();
		fPeak  = TMath::Max(fPeak,adc);
		fSum  += adc;
	}
	fSum *= gain;
	
	fStartTime = (fStartTimeSlot - tped)*tscale;
	fWidth     = (nslots - fNpres)*tscale;
#endif
#if 1
	next.Reset();
	for (islot = 0; islot < nslots; islot++) {
		adcp = (CDCRawADCDatum *)next();
		Double_t adc = ped - adcp->GetADC();
		fPeak  = TMath::Max(fPeak,adc);
		fSum  += adc;
	}
	fSum *= gain;
	
	fStartTime = (fStartTimeSlot - tped)*tscale;
	fWidth     = nslots*tscale;
#endif
}

Bool_t CDCCluster::IsSortable() const { return kTRUE; }

Int_t CDCCluster::Compare(TObject *obj)
{
	if (!fDone) Update();
	if (!((CDCCluster *)obj)->fDone) ((CDCCluster *)obj)->Update();
	switch (fSortBy) {
	    case 1: // sort by charge:  bigger one first:-)
   		if (fSum > ((CDCCluster *)obj)->fSum) return -1;
   		else if (fSum < ((CDCCluster *)obj)->fSum) return 1;
  		else return 0;
	    case 2: // sort by pulse height:  bigger one first:-)
   		if (fPeak > ((CDCCluster *)obj)->fPeak) return -1;
   		else if (fPeak < ((CDCCluster *)obj)->fPeak) return 1;
  		else return 0;
	    default: // sort by time: faster one first:-)
   		if (fStartTime < ((CDCCluster *)obj)->fStartTime) return -1;
   		else if (fStartTime > ((CDCCluster *)obj)->fStartTime) return 1;
  		else return 0;
  	}
}

void CDCCluster::SetSortBy(Char_t *opt)
{
	switch (opt[0]) {
		case 'C':	// sort by "Charge"
			fSortBy = 1;
			break;
		case 'P': 	// sort by "PulseHeight"
			fSortBy = 2;
			break;
		default: 	// sort by "Time"
			fSortBy = 0;
	}		
}

void CDCCluster::DebugPrint(const Char_t *opt)
{
	cerr.setf(ios::right);
	cerr.setf(ios::fixed);
	cerr	<< " (n,a)=(" << setw(1) << GetStationNo() << ","
			     << setw(1) << GetChannelNo() << ")"
   		<< " t = " << setw(4) << GetStartTime()
   		<< " w = " << setw(6) << setprecision(1) << GetWidth()
   		<< " Sum = " << setw(6) << setprecision(1) << GetSum()
    		<< " Pek = " << setw(5) << setprecision(1) << GetPeak()
    		<< " Ped = " << setw(5) << setprecision(1) << GetPedestal()
    		<< endl;
	cerr.unsetf(ios::right);
	cerr.unsetf(ios::fixed);

	if ( opt == "Detailed" ) {
		TIter next(this);
		CDCRawADCDatum *adc;
		while ((adc = (CDCRawADCDatum *)next())) {
			adc->DebugPrint();
		}
    	}
}

//_____________________________________________________________________
// --------------
// Cluster Buffer
// --------------

ClassImp(CDCClusterBuf)

CDCClusterBuf::CDCClusterBuf(CDCFADCCalibration *calib,
			     const char *name,const char *title, 
			     CDCClusterMaker *module) :
	JSFEventBuf(name, title, (JSFModule*)module),
	fDone(kFALSE), fUseVarPed(kTRUE), fNADCData(0), fADCData(0), fClusters(1), 
	fCalib(calib) {}

CDCClusterBuf::CDCClusterBuf(CDCClusterMaker *module,
			     CDCFADCCalibration *calib,
			     const char *name, const char *title) :
	JSFEventBuf(name, title, (JSFModule*)module),
	fDone(kFALSE), fUseVarPed(kTRUE), fNADCData(0), fADCData(0), fClusters(1), 
	fCalib(calib) {}

CDCClusterBuf::~CDCClusterBuf()
{
	fClusters.Delete();
}

void CDCClusterBuf::SetADCData(Len_t nadcdata, TClonesArray *adcdata)
{
	fDone     = kFALSE;
	fNADCData = nadcdata;
	fADCData  = adcdata;
	if (fClusters.GetEntries() > 0) fClusters.Delete();
}

void CDCClusterBuf::SetCalibrationData(CDCFADCCalibration *calib)
{
	fDone  = kFALSE;
	fCalib = calib;
}

Bool_t CDCClusterBuf::IsInitialized()  const
{
	if (fNADCData > 0) return kTRUE;
	else return kFALSE;
}

Int_t CDCClusterBuf::GetNclusters()
{
	if (!fDone) FindClusters(); 
	return fClusters.GetEntries();
}

TObjArray * CDCClusterBuf::GetClusters()
{
	if (!fDone) FindClusters();
	return &fClusters;
}

void CDCClusterBuf::UseVarPed()
{
	if (!IsInitialized()) {
		fUseVarPed = kTRUE;
	} else {
		if (!fUseVarPed) {
			fUseVarPed = kTRUE;
			TIter next(&fClusters);
			CDCCluster *clp;
			while ((clp = (CDCCluster *)next())) {
				clp->UseVarPed();
			}
		}
	}
}

void CDCClusterBuf::UseConstPed()
{
	if (!IsInitialized()) {
		fUseVarPed = kFALSE;
	} else {
		if (fUseVarPed) {
			fUseVarPed = kFALSE;
			TIter next(&fClusters);
			CDCCluster *clp;
			while ((clp = (CDCCluster *)next())) {
				clp->UseConstPed();
			}
		}
	}
}

void CDCClusterBuf::FindClusters()
{
	if (!IsInitialized()) {
		cerr << "ERROR  CDCClusterBuf::FindCluster" << endl
		     << "...... ADC data not yet initialized." << endl;
		return;
	} else if (fDone) {
		return;
	}
	
	fDone = kTRUE;
	CDCCluster *cur = new CDCCluster(fCalib,fUseVarPed);
	fClusters.Add(cur);
	
	TIter next(fADCData);
	CDCRawADCDatum *adcp;
	while ( (adcp = (CDCRawADCDatum *)next()) ) {
		if (! cur->Append(adcp)) {
			cur = new CDCCluster(fCalib,fUseVarPed);
			fClusters.Add(cur);
			cur->Append(adcp);
		}
	}
}

void CDCClusterBuf::DebugPrint(const Char_t *opt)
{
	cerr	<< fClusters.GetEntries()
		<< " clusters in this event." << endl;
	TIter next(&fClusters);
	CDCCluster *clp;
	Int_t n = 0;
	while ((clp = (CDCCluster *)next())) {
		cerr << setw(3) << ++n << " ";
		clp->DebugPrint(opt);
	}
}


//_____________________________________________________________________
// -------------
// Cluster Maker
// -------------

ClassImp(CDCClusterMaker)

CDCClusterMaker::CDCClusterMaker(CDCFADCCalibration *calib,
				 const char *name, const char *title)
       : JSFModule(name,title),
         fToDeleteCalib(calib ? kFALSE : kTRUE),
         fCalib(calib) {}

CDCClusterMaker::~CDCClusterMaker()
{
	if (fToDeleteCalib && fCalib) delete fCalib;
}

void CDCClusterMaker::SetCalibrationData(CDCFADCCalibration *calib)
{
	fCalib = calib;
	fToDeleteCalib = kFALSE;
	((CDCClusterBuf *)EventBuf())->SetCalibrationData(calib);
}

void CDCClusterMaker::UseVarPed()
{
	fUseVarPed = kTRUE;
	if (EventBuf()) ((CDCClusterBuf *)EventBuf())->UseVarPed();
}

void CDCClusterMaker::UseConstPed()
{
	fUseVarPed = kFALSE;
	if (EventBuf()) ((CDCClusterBuf *)EventBuf())->UseConstPed();
}

Bool_t CDCClusterMaker::Initialize()
{
	if (!fCalib) {
		fCalib = new CDCFADCCalibration();
		fToDeleteCalib = kTRUE;
	}
	fEventBuf = new CDCClusterBuf(this,fCalib);
	if (fUseVarPed) ((CDCClusterBuf *)fEventBuf)->UseVarPed();
	else		((CDCClusterBuf *)fEventBuf)->UseConstPed();
	
	return kTRUE;
}

Bool_t CDCClusterMaker::BeginRun(Int_t run)
{
	fCalib->Update();
	
	return kTRUE;
}

Bool_t CDCClusterMaker::Process(Int_t ev)
{
	CDCRawData    *raw    = (CDCRawData*)gJSF->FindModule("CDCRawData");
	CDCRawDataBuf *evt    = (CDCRawDataBuf *)raw->EventBuf();
	Int_t         nadc    = evt->GetNADCData();
	TClonesArray *adcdata = evt->GetADCData();

	CDCClusterBuf *clbufp = (CDCClusterBuf *)EventBuf();
	clbufp->SetADCData(nadc,adcdata);	
	clbufp->FindClusters();

	return kTRUE;
}
