#ifndef __CDCCLUSTERMAKER__
#define __CDCCLUSTERMAKER__

//*************************************************************************
//* ---------------------
//* CDCClusterMaker Class
//* ---------------------
//* (Description)
//*	This set of classes makes FADC clusters.
//* (Requires)
//*	TObjArray:
//*	JSFSteer:
//*	JSFModules:
//*	CDCRawAData:
//* (Provides)
//*	CDCClusterMaker		: steering class
//*	CDCClusterBuf		: buffer to store clusters
//*	CDCCluster		: cluster data class
//*	CDCFADCCalibration	: FADC calibration
//* (Update Record)
//*	1999/12/04  K.Fujii	Original version.
//*************************************************************************

#include <iostream.h>
#include <iomanip.h>
#include "TClass.h"
#include "TObjArray.h"

#include "JSFSteer.h"
#include "JSFModule.h"
#include "CDCRawData.h"

// ---------------------
// FADC Calibration Data
// ---------------------

class CDCFADCCalibration : public TObjArray {
private:
  class CalibDatum : public TObject {
     public:
     	CalibDatum(Double_t *ped  = 0, Double_t *gain   = 0, 
  		   Double_t *tped = 0, Double_t *tscale = 0, 
  		   Int_t last = 9999999) : fLastRun(last)
  	{
  	   Int_t i;
  	   if (!ped)    for (i=0; i<4; i++) fPedestal    [i] = 200.;
  	   else	     	for (i=0; i<4; i++) fPedestal    [i] = ped[i];
  	   if (!gain)   for (i=0; i<4; i++) fGain        [i] =   1.;
  	   else	     	for (i=0; i<4; i++) fGain        [i] = gain[i];
  	   if (!tped)   for (i=0; i<4; i++) fTimePedestal[i] =   0.;
  	   else	     	for (i=0; i<4; i++) fTimePedestal[i] = tped[i];
  	   if (!tscale) for (i=0; i<4; i++) fTimeScale   [i] =   2.;
  	   else         for (i=0; i<4; i++) fTimeScale   [i] = tscale[i];
  	}
  	Int_t 	 GetLastRun() { return fLastRun; }	
	Double_t GetPedestal    (Int_t n, Int_t a) { return fPedestal[a];    }
	Double_t GetGain        (Int_t n, Int_t a) { return fGain[a];        }
	Double_t GetTimePedestal(Int_t n, Int_t a) { return fTimePedestal[a];}
	Double_t GetTimeScale   (Int_t n, Int_t a) { return fTimeScale[a];   }
     private:
	Int_t	 fLastRun;		//! Last Run Number
	Double_t fPedestal[4];		//! Pedestal
	Double_t fGain[4];		//! Gain
	Double_t fTimePedestal[4];	//! Time Pedestal
	Double_t fTimeScale[4];		//! Time Scale
  };
  
public:
  CDCFADCCalibration(Double_t *ped  = 0, Double_t *gain   = 0, 
  		     Double_t *tped = 0, Double_t *tscale = 0,
  		     Int_t last = 9999999);
  CDCFADCCalibration(Char_t * calibfile);
  virtual ~CDCFADCCalibration();
  
  Double_t GetPedestal    (Int_t n, Int_t a); 
  Double_t GetGain        (Int_t n, Int_t a);
  Double_t GetTimePedestal(Int_t n, Int_t a);
  Double_t GetTimeScale   (Int_t n, Int_t a);

  void Update();
  
private:
  CalibDatum *fCurr;	//! pointer to current calib datum

  ClassDef(CDCFADCCalibration, 1) 	// CDC FADC Calibration
};


// -------------
// Cluster Datum
// -------------

class CDCCluster : public TObjArray {
public:
  CDCCluster(CDCFADCCalibration *calib = 0, Bool_t varped = kTRUE);
  CDCCluster(CDCCluster &cl);
  virtual ~CDCCluster();

  void	SetNpeds(Int_t n);

  inline Int_t 	GetStationNo()     const  { return fStationNo;     }
  inline Int_t 	GetChannelNo()     const  { return fChannelNo;     }
  inline Int_t 	GetStartTimeSlot() const  { return fStartTimeSlot; }
  inline Int_t 	GetLastTimeSlot()  const  { return fLastTimeSlot;  }
  Double_t 	GetSum()	{ if (!fDone) Update(); return fSum;       }
  Double_t	GetPeak()	{ if (!fDone) Update(); return fPeak;      }
  Double_t 	GetVarPedestal(){ if (!fDone) Update(); return fPedestal;  }
  Double_t	GetStartTime()	{ if (!fDone) Update(); return fStartTime; }
  Double_t	GetWidth()	{ if (!fDone) Update(); return fWidth; 	   }
  Double_t 	GetPedestal();

  void		UseVarPed();
  void		UseConstPed();
  Bool_t 	Append(CDCRawADCDatum *adc);

  Bool_t 	IsSortable() const;
  Int_t  	Compare(TObject *obj);
  static void	SetSortBy(Char_t *opt = "Time");

  void 		DebugPrint(const Char_t *opt = "Brief");

private:
  void		Update();

private:
  Bool_t	fDone;			// (1,0) = (cluster done, not yet)
  Bool_t	fUseVarPed;		// (1,0) = (var ped, const ped)
  Int_t	   	fStationNo;		// Station number
  Int_t    	fChannelNo;		// Channel number
  Int_t 	fStartTimeSlot;		// start time slot
  Int_t 	fLastTimeSlot;		// last  time slot
  Double_t 	fSum;			// ADC sum
  Double_t 	fPeak;			// ADC peak
  Double_t 	fPedestal;		// ADC pedestal
  Double_t	fStartTime;		// start time (ns)
  Double_t	fWidth;			// width (ns)
  Int_t		fNpeds;			// No. slots for pedestal calc.
  static Int_t	fNpres;			// No. pre-slots
  static Int_t	fNposts;		// No. post-slots
  static Int_t	fSortBy;		// (0,1,2) = by (Time, Peak, Charge)

  CDCFADCCalibration *fCalib;		// pointer to calibration data object

  ClassDef(CDCCluster, 1) 	// CDC Cluster
};

// --------------
// Cluster Buffer
// --------------
class CDCClusterMaker;

class CDCClusterBuf : public JSFEventBuf {
friend class CDCClusterMaker;
public:
  CDCClusterBuf(CDCFADCCalibration *calib=0,
  		const char *name="CDCClusterBuf",
		const char *title="CDC cluster data", 
		CDCClusterMaker *mod=0);
  CDCClusterBuf(CDCClusterMaker *mod, CDCFADCCalibration *calib=0,
		const char *name="CDCClusterBuf",
		const char *title="CDC cluster data");
  virtual ~CDCClusterBuf();

  void		SetADCData(Len_t nadcdata, TClonesArray *adcdata);
  void		SetCalibrationData(CDCFADCCalibration *calib);

  Bool_t	IsInitialized() const;	// inquires if input data set
  Int_t		GetNclusters();		// returns No. clusters
  TObjArray    *GetClusters();		// returns array of clusters
  void		UseVarPed();		// use variable pedestal
  void		UseConstPed();		// use constant pedestal

  void 		DebugPrint(const Char_t *opt = "Brief");

private:
  void		FindClusters();		// finds clusters

private:
  Bool_t	      fDone;	 	// (1,0) = (clustering done, not yet)
  Bool_t	      fUseVarPed;	// (1,0) = (var ped, const ped)
  Len_t		      fNADCData; 	// No. ADC data units
  TClonesArray	     *fADCData;	 	// Array of ADC Data
  TObjArray	      fClusters; 	// array of clusters
  CDCFADCCalibration *fCalib;	 	// pointer to calibration data object

  ClassDef(CDCClusterBuf, 1) // CDC cluster data buffer
};


// -------------
// Cluster Maker
// -------------

class CDCClusterMaker : public JSFModule {
public:
  CDCClusterMaker(CDCFADCCalibration *calib = 0,
		  const Char_t *name="CDCClusterMaker",
		  const Char_t *title="CDC Cluster Maker");
  virtual ~CDCClusterMaker();

  void 		SetCalibrationData(CDCFADCCalibration *calib);
  void		UseVarPed();
  void 		UseConstPed();

  Bool_t 	Initialize();		// job initialization
  Bool_t 	BeginRun(Int_t run);	// process Begin Run record
  Bool_t 	Process(Int_t  ev);	// process event
  
private:
  Bool_t	      fUseVarPed;	// (1,0) = (var ped, const ped)
  Bool_t	      fToDeleteCalib;	// delete Calib in dtor or not
  CDCFADCCalibration *fCalib;	 	// pointer to calibration data object
  
  ClassDef(CDCClusterMaker, 1) 	// CDC cluster maker
};

#endif



