//*************************************************************************
//* --------------------
//* CDCPadHitMaker Class
//* --------------------
//* (Description)
//*	This set of classes makes pad hits.
//* (Requires)
//*	TObjArray:
//*	JSFSteer:
//*	JSFModules:
//*	CDCCluster:
//*	CDCClusterBuf:
//*	CDCClusterMaker:
//* (Provides)
//*	CDCPadHitMaker		: steering class
//*	CDCPadHitBuf		: hit data buffer class
//*	CDCPadHit		: hit data class
//*	CDCPadCalibration	: Pad calibration
//* (Update Record)
//*	1999/12/07  K.Fujii	Original version.
//*************************************************************************

#include "CDCPadHitMaker.h"

//_____________________________________________________________________
// --------------------
// Pad Calibration Data
// --------------------

ClassImp(CDCPadCalibration)

CDCPadCalibration::CDCPadCalibration(Double_t xscale, Double_t width, 
  		    Double_t vdrift, Int_t last)
{
  	CalibDatum *datum = new CalibDatum(xscale,width,vdrift,last);
  	Add(datum);
  	fCurr = datum;
}

CDCPadCalibration::CDCPadCalibration(Char_t * calibfile) 
{
	cout << " Going to open a calib file " << calibfile << endl;
	FILE *fd = fopen(calibfile,"r");
	cout << " Opened " << calibfile << endl;
	if (fd == 0) {
		Error("CDCPadCalibration","Failed to open %s.",calibfile);
		return;
        }
        Int_t last = 99999999;
        Double_t xscale, width, vdrift;
        while ((fscanf(fd, "%d %lg %lg %lg",
        	       &last, &xscale, &width, &vdrift)) != EOF) {
	   CalibDatum *datum 
	   	= new CalibDatum(xscale,width,vdrift,last);
	   Add(datum);
	}
	fclose(fd);
}

CDCPadCalibration::~CDCPadCalibration() { Delete(); }

void CDCPadCalibration::Update()
{
  	Int_t run = gJSF->GetRunNumber();
  	TIter next(this);
  	CalibDatum *datum;
  	while ((datum = (CalibDatum *)next())) {
  		fCurr = datum;
  		if (run <= fCurr->GetLastRun()) break;
  	}	
}


//_____________________________________________________________________
// -------------
// Pad Hit Datum
// -------------

ClassImp(CDCPadHit)

CDCPadHit::CDCPadHit(CDCPadCalibration *calib) : 
	fDone(kFALSE), fPadSum(0.), fAnode(0),
	fPadLeft(0), fPadCenter(0), fPadRight(0), fX(0.), fZ(0.),
	fCalib(calib) {}

CDCPadHit::CDCPadHit(CDCPadCalibration *calib, CDCCluster *anode, 
		     CDCCluster *padl, CDCCluster *padc, CDCCluster *padr) :
  	fDone(kFALSE), fPadSum(0.), fAnode(anode), 
  	fPadLeft(padl), fPadCenter(padc), fPadRight(padr), fX(0.), fZ(0.),
  	fCalib(calib) {}

CDCPadHit::CDCPadHit(CDCPadHit & h) : 
	fDone(h.fDone), fPadSum(h.fPadSum), fAnode(h.fAnode), 
	fPadLeft(h.fPadLeft), fPadCenter(h.fPadCenter), fPadRight(h.fPadRight),
	fX(h.fX), fZ(h.fZ), fCalib(h.fCalib) {}

CDCPadHit::~CDCPadHit() {}


Double_t CDCPadHit::GetSum(const Char_t c)
{
	if (!fDone) Update();
	switch (c) {
		case 'a' : return fAnode->GetSum();
		case 'p' : return fPadSum;
		case 'l' : return fPadLeft->GetSum();
		case 'c' : return fPadCenter->GetSum();
		case 'r' : return fPadRight->GetSum();
		default  :
			cerr 	<< "???? CDCPadHit::GetSum ???" << endl
				<< " type = " << c 
				<< " not supported!" << endl
				<< " Default to \'p\'" << endl;
			return fPadSum;
	}
}

Bool_t CDCPadHit::IsSortable() const { return kTRUE; }

Int_t CDCPadHit::Compare(TObject *obj)
{
	if (!fDone) Update();
	Double_t tim  = fAnode->GetStartTime();
	Double_t timo = ((CDCPadHit *)obj)->fAnode->GetStartTime();
   	if (tim < timo) return -1;
   	else if (tim > timo) return 1;
   	else return 0;
}

void CDCPadHit::DebugPrint(const Char_t *opt)
{
	if (!fDone) Update();
	Double_t pl = fPadLeft->GetSum();
	Double_t pc = fPadCenter->GetSum();
	Double_t pr = fPadRight->GetSum();
	cerr.setf(ios::right);
	cerr.setf(ios::fixed);
	cerr	<< " P_sum = " << setw(6) << setprecision(1) << fPadSum
		<< " R_L = " << setw(5) << setprecision(3) << pl/pc
		<< " R_R = " << setw(5) << setprecision(3) << pr/pc
		<< " x = " << setw(5) << setprecision(3) << fX		
		<< " z = " << setw(5) << setprecision(2) << fZ << endl;

	if (strstr(opt,"Detailed")) {
		cerr 	<< "    "
			<< " P_A = " << setw(6) << setprecision(1) 
				  	  << fAnode->GetSum()
			<< " P_L = " << setw(6) << setprecision(1) << pl
			<< " P_C = " << setw(6) << setprecision(1) << pc
			<< " P_R = " << setw(6) << setprecision(1) << pr
			<< endl;
    	}
	cerr.unsetf(ios::right);
	cerr.unsetf(ios::fixed);
}

void CDCPadHit::Update()
{
	if (fDone) return;
	fDone = kTRUE;
	fPadSum = fPadLeft->GetSum() 
		+ fPadCenter->GetSum() 
		+ fPadRight->GetSum();
	CalculatePosition(fAnode,fPadLeft,fPadCenter,fPadRight,fX,fZ);
}

void CDCPadHit::CalculatePosition(CDCCluster *a,  CDCCluster *l,
  				  CDCCluster *c, CDCCluster *r,
  				  Double_t &x, Double_t &z)
{
	Double_t calib  = fCalib->GetXScale();
	Double_t width  = fCalib->GetPadWidth();
	Double_t vdrift = fCalib->GetVdrift();

	// Double_t pa = a->GetSum();
	Double_t pl = l->GetSum();
	Double_t pc = c->GetSum();
	Double_t pr = r->GetSum();
	x = calib*width*(pr - pl)/(pl + pc + pr);
	z = a->GetStartTime() * vdrift;
}

//_____________________________________________________________________
// --------------
// Pad Hit Buffer
// --------------

ClassImp(CDCPadHitBuf)

CDCPadHitBuf::CDCPadHitBuf(CDCPadCalibration *calib,
			   const char *name,const char *title, 
			   CDCPadHitMaker *module) :
	JSFEventBuf(name, title, (JSFModule*)module),
	fDone(kFALSE), fClusterData(0), fHits(1), 
	fCalib(calib) {}

CDCPadHitBuf::CDCPadHitBuf(CDCPadHitMaker *module,
			   CDCPadCalibration *calib,
			   const char *name, const char *title) :
	JSFEventBuf(name, title, (JSFModule*)module),
	fDone(kFALSE), fClusterData(0), fHits(1), 
	fCalib(calib) {}

CDCPadHitBuf::~CDCPadHitBuf()
{
	fHits.Delete();
}

void CDCPadHitBuf::SetClusterData(CDCClusterBuf *clbuf)
{
	fDone        = kFALSE;
	fClusterData = clbuf;
	if (fHits.GetEntries() > 0) fHits.Delete();
}

void CDCPadHitBuf::SetCalibrationData(CDCPadCalibration *calib)
{
	fDone  = kFALSE;
	fCalib = calib;
}

Bool_t CDCPadHitBuf::IsInitialized()  const
{
	if (fClusterData) return kTRUE;
	else return kFALSE;
}

Int_t CDCPadHitBuf::GetNhits()
{
	if (!fDone) MakeHits(); 
	return fHits.GetEntries();
}

TObjArray * CDCPadHitBuf::GetHits()
{
	if (!fDone) MakeHits();
	return &fHits;
}

void CDCPadHitBuf::MakeHits()
{
	if (!IsInitialized()) {
		cerr << "ERROR  CDCPadHitBuf::MakeHits" << endl
		     << "...... cluster data not yet initialized." << endl;
		return;
	} else if (fDone) {
		return;
	}
	
	fDone = kTRUE;

	TObjArray pap(1), plp(1), pcp(1), prp(1);
	TIter next(fClusterData->GetClusters());
	CDCCluster *clp;
	while ((clp = (CDCCluster *)next())) {
		Int_t n = clp->GetStationNo();
		Int_t a = clp->GetChannelNo();
		if (n) {
			cerr 	<< "???? CDCPadHitBuf::MakeHits ????"
				<< " n = " << n 
				<< " unexpected!" << endl;
			break;
		}
		switch (a) {
			case 0:
				plp.Add(clp);
				break;
			case 1:
				pcp.Add(clp);
				break;
			case 2:
				prp.Add(clp);
				break;
			case 3:
				pap.Add(clp);
				break;
			default:
				cerr 	<< "???? CDCPadHitBuf::MakeHits ????"
				<< " a = " << a 
				<< " unexpected!" << endl;
		}
	}
	
	if (pap.GetEntries() < 2 || plp.GetEntries() < 2 ||
	    pcp.GetEntries() < 2 || prp.GetEntries() < 2) return;

	Int_t i, n; // remove small hits, leaving only the biggest two.
	CDCCluster::SetSortBy("Charge");
	n=pap.GetEntries(); pap.Sort(); for (i=2; i<n; i++) pap.RemoveAt(i);
	n=plp.GetEntries(); plp.Sort(); for (i=2; i<n; i++) plp.RemoveAt(i);
	n=pcp.GetEntries(); pcp.Sort(); for (i=2; i<n; i++) pcp.RemoveAt(i);
	n=prp.GetEntries(); prp.Sort(); for (i=2; i<n; i++) prp.RemoveAt(i);

	CDCCluster *Dummyp = new CDCCluster(); 	// A Stupid trick:
	pap.Add(Dummyp); pap.Remove(Dummyp);	// need to touch the these
	plp.Add(Dummyp); plp.Remove(Dummyp);	// TObjArrays, otherwise
	pcp.Add(Dummyp); pcp.Remove(Dummyp);	// ROOT will not resort
	prp.Add(Dummyp); prp.Remove(Dummyp);	// them below.
	delete Dummyp;

	CDCCluster::SetSortBy("Time");	
	pap.Sort(); 				// resort them in ascending
	plp.Sort();				// order of start times.
	pcp.Sort();
	prp.Sort();
	
	for (i = 0; i < 2; i++) {
		CDCPadHit *padhit = new CDCPadHit(fCalib, 
						  (CDCCluster *)pap[i], 
						  (CDCCluster *)plp[i],
						  (CDCCluster *)pcp[i],
						  (CDCCluster *)prp[i]);
		fHits.Add(padhit);
	}
}

void CDCPadHitBuf::DebugPrint(const Char_t *opt)
{
	cerr	<< fHits.GetEntries()
		<< " pad hits in this event:";
	if (fHits.GetEntries() > 1) {
		Double_t dx = ((CDCPadHit *)fHits.At(0))->X()
			    - ((CDCPadHit *)fHits.At(1))->X();
		cerr 	<< " dx = " << setprecision(3) << dx;
	}
	cerr 	<< endl;
	
	TIter next(&fHits);
	CDCPadHit *php;
	Int_t n = 0;
	while ((php = (CDCPadHit *)next())) {
		cerr << setw(3) << ++n << " ";
		php->DebugPrint(opt);
	}
}


//_____________________________________________________________________
// -------------
// Cluster Maker
// -------------

ClassImp(CDCPadHitMaker)

CDCPadHitMaker::CDCPadHitMaker(CDCPadCalibration *calib,
			       const char *name, const char *title)
       : JSFModule(name,title), 
         fToDeleteCalib(calib ? kFALSE : kTRUE), fCalib(calib) {}

CDCPadHitMaker::~CDCPadHitMaker()
{
	if (fToDeleteCalib && fCalib) delete fCalib;
}

void CDCPadHitMaker::SetCalibrationData(CDCPadCalibration *calib)
{
	fCalib = calib;
	fToDeleteCalib = kFALSE;
	((CDCPadHitBuf *)EventBuf())->SetCalibrationData(calib);
}

Bool_t CDCPadHitMaker::Initialize()
{
	if (!fCalib) {
		fCalib = new CDCPadCalibration();
		fToDeleteCalib = kTRUE;
	}
	fEventBuf = new CDCPadHitBuf(this,fCalib);
	
	return kTRUE;
}

Bool_t CDCPadHitMaker::BeginRun(Int_t run)
{
	fCalib->Update();
	
	return kTRUE;
}

Bool_t CDCPadHitMaker::Process(Int_t ev)
{
	CDCClusterMaker *clmp
		= (CDCClusterMaker*)gJSF->FindModule("CDCClusterMaker");
	CDCClusterBuf *clbufp = (CDCClusterBuf *)clmp->EventBuf();

	CDCPadHitBuf  *phbufp = (CDCPadHitBuf *)EventBuf();
	phbufp->SetClusterData(clbufp);	
	phbufp->MakeHits();

	return kTRUE;
}

