#ifndef __CDCPADHITMAKER__
#define __CDCPADHITMAKER__

//*************************************************************************
//* --------------------
//* CDCPadHitMaker Class
//* --------------------
//* (Description)
//*	This set of classes makes pad hits.
//* (Requires)
//*	TObjArray:
//*	JSFSteer:
//*	JSFModules:
//*	CDCCluster:
//*	CDCClusterBuf:
//*	CDCClusterMaker:
//* (Provides)
//*	CDCPadHitMaker		: steering class
//*	CDCPadHitBuf		: hit data buffer class
//*	CDCPadHit		: hit data class
//*	CDCPadCalibration	: Pad calibration
//* (Update Record)
//*	1999/12/07  K.Fujii	Original version.
//*************************************************************************

#include "CDCClusterMaker.h"

// --------------------
// Pad Calibration Data
// --------------------

class CDCPadCalibration : public TObjArray {
private:
  class CalibDatum : public TObject {
     public:
     	CalibDatum(Double_t xscale = 1.36, Double_t width  = 5., 
  		   Double_t vdrift = 0.07,  
  		   Int_t last = 9999999) : fLastRun(last)
  	{
		fXScale   = xscale;
		fPadWidth = width;
		fVdrift   = vdrift;
  	}
  	Int_t 	 GetLastRun  () { return fLastRun; }	
	Double_t GetXScale   () { return fXScale;  } 	// x scale calib
	Double_t GetPadWidth () { return fPadWidth;}	// 5mm
	Double_t GetVdrift   () { return fVdrift;  } 	// 10microns/ns
     private:
	Int_t	 fLastRun;		//! Last Run Number
	Double_t fXScale;		//! X scale
	Double_t fPadWidth;		//! Pad width
	Double_t fVdrift;		//! drift velocity
  };
  
public:
  CDCPadCalibration(Double_t xscale = 1.36, Double_t width  = 5., 
  		    Double_t vdrift = 0.07,  
  		    Int_t last = 9999999);
  CDCPadCalibration(Char_t * calibfile);

  virtual ~CDCPadCalibration();
  Int_t	   GetLastRun  () { return fCurr->GetLastRun(); }	
  Double_t GetXScale   () { return fCurr->GetXScale();  }
  Double_t GetPadWidth () { return fCurr->GetPadWidth();}
  Double_t GetVdrift   () { return fCurr->GetVdrift();  }

  void Update();
  
private:
  CalibDatum *fCurr;	//! pointer to current calib datum

  ClassDef(CDCPadCalibration, 1) 	// CDC Pad Calibration
};



// -------------
// Pad Hit Datum
// -------------

class CDCPadHit : public TObject {
friend class CDCPadHitMaker;
friend class CDCPadHitBuf;
public:
  CDCPadHit(CDCPadCalibration *calib = 0);
  CDCPadHit(CDCPadCalibration *calib, CDCCluster *anode, 
  	    CDCCluster *padl, CDCCluster *padc, CDCCluster *padr);
  CDCPadHit(CDCPadHit &cl);
  virtual ~CDCPadHit();

  Double_t	X() { if (!fDone) Update(); return fX; }
  Double_t	Z() { if (!fDone) Update(); return fZ; }
  Double_t 	GetSum(const Char_t c = 'p');
  
  Bool_t	IsSortable() const;
  Int_t		Compare(TObject *obj);

  void 		DebugPrint(const Char_t *opt = "Brief");

private:
  void		Update();
  virtual void	CalculatePosition(CDCCluster *a, CDCCluster *l,
  				  CDCCluster *c, CDCCluster *r,
  				  Double_t &x, Double_t &z);

private:
  Bool_t	fDone;			// (1,0) = (cluster done, not yet)
  Double_t 	fPadSum;		// pad ADC sum
  CDCCluster   *fAnode;			// pointer to left pad
  CDCCluster   *fPadLeft;		// pointer to left pad
  CDCCluster   *fPadCenter;		// pointer to center pad
  CDCCluster   *fPadRight;		// pointer to right pad
  Double_t	fX;			// x position (mm)
  Double_t	fZ;			// y position (mm)

  CDCPadCalibration  *fCalib;		// pointer to pad calibration data

  ClassDef(CDCPadHit, 1) 	// CDC Pad Hit
};

// ----------
// Hit Buffer
// ----------
class CDCPadHitMaker;

class CDCPadHitBuf : public JSFEventBuf {
friend class CDCPadHitMaker;
public:
  CDCPadHitBuf(CDCPadCalibration *calib = 0,
  	       const char *name="CDCPadHitBuf",
	       const char *title="CDC pad hit data", 
	       CDCPadHitMaker *mod=0);
  CDCPadHitBuf(CDCPadHitMaker *mod,
  	       CDCPadCalibration *calib,
	       const char *name="CDCPadHitBuf",
	       const char *title="CDC pad hit data");
  virtual ~CDCPadHitBuf();

  void		SetClusterData(CDCClusterBuf *clbuf);
  void		SetCalibrationData(CDCPadCalibration *calib);

  Bool_t	IsInitialized() const;	// inquires if input data set
  Int_t		GetNhits();		// returns No. hits
  TObjArray    *GetHits();		// returns array of hits

  void 		DebugPrint(const Char_t *opt = "Brief");

private:
  void		MakeHits();		// make hits

private:
  Bool_t	      fDone;	 	// (1,0) = (clustering done, not yet)
  CDCClusterBuf	     *fClusterData;	// Array of clusters
  TObjArray	      fHits; 		// array of hits
  CDCPadCalibration  *fCalib;		// pointer to pad calibration data

  ClassDef(CDCPadHitBuf, 1) // CDC pad hit data buffer
};


// -------------
// Pad Hit Maker
// -------------

class CDCPadHitMaker : public JSFModule {
public:
  CDCPadHitMaker(CDCPadCalibration *calib = 0,
  		 const Char_t *name="CDCPadHitMaker",
		 const Char_t *title="CDC Pad Hit Maker");
  virtual ~CDCPadHitMaker();

  void 		SetCalibrationData(CDCPadCalibration *calib);

  Bool_t 	Initialize();		// job initialization
  Bool_t 	BeginRun(Int_t run);	// process Begin Run record
  Bool_t 	Process(Int_t ev);	// process event

private:
  Bool_t	      fToDeleteCalib;	// delete Calib in dtor or not
  CDCPadCalibration  *fCalib;		// pointer to pad calibration data
  
  ClassDef(CDCPadHitMaker, 1) 	// CDC Pad Hit Maker
};

#endif




