//*************************************************************************
//* ----------------
//* CDCRawData Class
//* ----------------
//* (Description)
//*	This set of classes reads CDC raw data for Lorentz Angle 
//*	measurments and unpacks them into root objects.
//*	When a root file is opened in the read/write mode, the root
//*	objects are stored in it.  If the root file is read-only,
//*	the data are obtained from it.
//*	The package contains the following classes:
//*	The following parameters can be specified in "jsf.conf". 
//*
//*	    CDCRawData.RFNFormat:       printf format to construct file name
//*	    CDCRawData.DoPlots:         (TRUE/FASLE)
//* (Requires)
//*	TObject:
//*	TClonesArray:
//*	CDCUniversalRecordHeader:
//*	CDCRawADCDatum:
//* (Provides)
//*	CDCRawDataBuf		: raw data buffer
//*	CDCBeginRunBuf		: begin run buffer
//*	CDCEndRunBuf		: end run buffer
//*	CDCRawData			: steering class
//* (Update Record)
//*	1999/06/02  K.Fujii	Original version.
//*************************************************************************

#include "JSFSteer.h"
#include "CDCRawData.h"
#include "TRandom.h"
#include <iostream.h>

ClassImp(CDCBeginRunBuf)
ClassImp(CDCRawDataBuf)
ClassImp(CDCEndRunBuf)
ClassImp(CDCRawData)

static const UInt_t kNADCDataMax = 120000;

//___________________________________________________________________________
// ----------------
// Begin Run Buffer
// ----------------
//
CDCBeginRunBuf::CDCBeginRunBuf()
{
   fDateTime = 0;
   fComment  = 0;
}

//___________________________________________________________________________
CDCBeginRunBuf::~CDCBeginRunBuf()
{
//   if ( fMap )     delete fMap;
//   if ( fThresh )  delete fThresh;
   if ( fComment ) delete fComment;
}

//___________________________________________________________________________
Bool_t CDCBeginRunBuf::ReadFile(FILE *fd) {
    fHeader.ReadFile(fd);
    if ( fHeader.GetRecordType() != kRecTypeBRR ) return kFALSE;
    Len_t   nw   = fHeader.GetDataLength();
    cerr << "nw = " << nw << endl;
    BRR_t  *data = new BRR_t[nw];
    Len_t   ne   = fread(data, sizeof(BRR_t), nw, fd);
    if ( ne != nw ) {
        Error("BeginRunBuf","Failed to read BRR: nw =%d",ne);
        delete data;
        return kFALSE;
    }
    Unpack(data);
    delete data;
    return kTRUE;
}

//___________________________________________________________________________
void CDCBeginRunBuf::Unpack(BRR_t data[])
{
}

//___________________________________________________________________________
// ---------------
// Raw Data Buffer
// ---------------
//
CDCRawDataBuf::CDCRawDataBuf(const char *name, const char *title, 
	CDCRawData *module) : JSFEventBuf(name, title, (JSFModule*)module),
			      fNADCData(0), fADCData(0) {}

//___________________________________________________________________________
CDCRawDataBuf::CDCRawDataBuf(CDCRawData *module, const char *name, 
	const char *title) : JSFEventBuf(name, title, (JSFModule*)module),
			     fNADCData(0), fADCData(0) {}

//___________________________________________________________________________
CDCRawDataBuf::~CDCRawDataBuf()
{
    if ( fADCData ) delete fADCData;
}

//___________________________________________________________________________
Bool_t CDCRawDataBuf::ReadFile(FILE *fd)
{
    if ( !fHeader.ReadFile(fd) ) return kFALSE;
    if ( fHeader.GetRecordType() != kRecTypeADC ) return kFALSE;
    Len_t   nw   = fHeader.GetDataLength();
    ADC_t  *data = new ADC_t[nw];
    Len_t   ne   = fread(data, sizeof(ADC_t), nw, fd);
    if ( ne != nw ) {
        Error("RawDataBuf","Failed to read Event Data: nw =%d",ne);
        delete data;
        return kFALSE;
    }
    CDCRawADCDatum::SetFormat(fHeader);
    fNADCData = nw/CDCRawADCDatum::GetRawLength();
    Unpack(data);
    delete data;
    return kTRUE;
}

//___________________________________________________________________________
void CDCRawDataBuf::Unpack(ADC_t data[])
{
    if ( !fADCData ) fADCData = new TClonesArray("CDCRawADCDatum",fNADCData);
    TClonesArray & adcdata = *fADCData;
    Len_t ladc = CDCRawADCDatum::GetRawLength();
    for ( Int_t i = 0; i < fNADCData; i++ ) {
        new(adcdata[i]) CDCRawADCDatum(&data[ladc*i]);
    }
}
//___________________________________________________________________________
void CDCRawDataBuf::Streamer(TBuffer &R__b)
{
   // Stream an object of class CDCRawDataBuf.

   if (R__b.IsReading()) {
      Version_t R__v = R__b.ReadVersion(); if (R__v) { }
      JSFEventBuf::Streamer(R__b);
      fHeader.Streamer(R__b);
      R__b >> fNADCData;
      if (!fADCData) fADCData = new TClonesArray("CDCRawADCDatum",fNADCData);
      TClonesArray & adcdata = *fADCData;
      cerr << "fNADCData = " << fNADCData << endl;
      for ( Int_t i = 0; i < fNADCData; i++ ) {
          new(adcdata[i]) CDCRawADCDatum();
          ((CDCRawADCDatum *)adcdata.UncheckedAt(i))->Streamer(R__b);
      }
   } else {
      R__b.WriteVersion(CDCRawDataBuf::IsA());
      JSFEventBuf::Streamer(R__b);
      fHeader.Streamer(R__b);
      R__b << fNADCData;
      TClonesArray & adcdata = *fADCData;
      for ( Int_t i = 0; i < fNADCData; i++ ) {
          ((CDCRawADCDatum *)adcdata.UncheckedAt(i))->Streamer(R__b);
      }
   }
}

//___________________________________________________________________________
// --------------
// End Run Buffer
// --------------
//
Bool_t CDCEndRunBuf::ReadFile(FILE *fd) {
    fHeader.ReadFile(fd);
    if ( fHeader.GetRecordType() != kRecTypeERR ) return kFALSE;
    Len_t  nw   = fHeader.GetDataLength();
    ERR_t *data = new ERR_t[nw];
    Len_t  ne = fread(data, sizeof(ERR_t), nw, fd);
    if ( ne != nw ) {
        Error("BeginRunBuf","Failed to read ERR: nw =%d",ne);
        delete data;
        return kFALSE;
    }
    Unpack(data);
    delete data;
    return kTRUE;
}
//___________________________________________________________________________
void CDCEndRunBuf::Unpack(ERR_t data[])
{
}

//___________________________________________________________________________
// -----------------
// Raw Data Analysis
// -----------------
//
CDCRawData::CDCRawData(const char *name, const char *title)
       : JSFModule(name,title)
{
    cerr << " constructor for CDCRawData is called..." << endl;
    fBeginBuf = new CDCBeginRunBuf();
    fEventBuf = new CDCRawDataBuf(this);

    Len_t len = strlen( gJSF->Env()->GetValue("CDCRawData.RFNFormat",
                         "/proj/cdc/data/lorentz/1999/run%4.4d.dat") );
    fRFNFormat = new Char_t[len+1];
    strcpy( fRFNFormat,  gJSF->Env()->GetValue("CDCRawData.RFNFormat",
                         "/proj/cdc/data/lorentz/1999/run%4.4d.dat") );
    fd = NULL;
    SetBufferSize(2000000);  // buffer size for event data.
}

//___________________________________________________________________________
CDCRawData::~CDCRawData()
{
    cerr << " CDCRawData will be deleted...fEventBuf is " 
    	 << (Int_t)fEventBuf << endl;
    delete fBeginBuf;
    delete fEventBuf;
    delete fRFNFormat;
    if(!fd) fclose(fd);
}

//___________________________________________________________________________
Bool_t CDCRawData::BeginRun(Int_t nrun)
{
    cout << " BeginRun is called for run " << nrun << endl;
        
    // Open Raw Data Root File
   
    if ( ! fFile->IsWritable() ) {
        return kTRUE;
    }

    // Open Raw Data File
    
    Len_t ln = strlen(fRFNFormat);
    Char_t *infile = new Char_t[ln+10];
    sprintf(infile,fRFNFormat,nrun);
    cout << " Going to open a file " << infile << endl;

    fd = fopen(infile,"rb");

    if ( fd == NULL ) {
        Error("BeginRun","Failed to open %s.",infile);
        delete infile;
        return kFALSE;
    }
    delete infile;
    
    // Read and Unpack Begin Run Record
    
    cout << " Going to unpack BRR " << endl;
    if ( ! fBeginBuf->ReadFile(fd) ) return kFALSE;
    
    Int_t irun = fBeginBuf->GetHeader()->GetRunNumber();
    fBeginBuf->GetHeader()->DebugPrint();
    cout << " BRR unpacked " << endl;
  
    // output standard information
    
    SetModuleStatus(kBeginRun);
    SetRunNumber(irun);
    ConfDirectory();
    
    // Save begin run record
    
    TDirectory *last = gDirectory;
    Char_t dir[100];
    sprintf(dir,"/conf/begin%5.5i",nrun);
    fFile->cd(dir);
    cout << " current file name is " << fFile->GetPath() << endl;

    gDirectory->pwd();
    fBeginBuf->Write("CDCBeginRunBuf");
    last->cd();

    return kTRUE;
}

//___________________________________________________________________________
Bool_t CDCRawData::Process(Int_t ev)
{
    CDCRawDataBuf *buf = (CDCRawDataBuf*)fEventBuf;
    if ( fFile->IsWritable() ) {
REPEAT:
        while (!buf->ReadFile(fd)) {
            Int_t id = buf->GetHeader()->GetRecordType();
            switch (id) {
                case kRecTypeERR: {
                    cout << "ERR encountered " << endl;
                    SetModuleStatus(kEndRun);
                    return kFALSE;
                }
                case -9999: {
                    cout << "EOF encountered " << endl;
                    return kFALSE;
                }
                default: {
                    cout << " Obtained record type is " << id << endl;
                }
            }
        }
    
        if (gJSF->GetRunNumber() != buf->GetHeader()->GetRunNumber()) {
            cerr << "????? ERROR ????? Run number inconsistency" << endl;
            buf->GetHeader()->DebugPrint();
            goto REPEAT;
        }
    }
    buf->GetHeader()->DebugPrint();

    return kTRUE;
}





