#ifndef __CDCRAWDATA__
#define __CDCRAWDATA__

//*************************************************************************
//* ----------------
//* CDCRawData Class
//* ----------------
//* (Description)
//*	This set of classes reads CDC raw data for Lorentz Angle 
//*	measurments and unpacks them into root objects.
//*	When a root file is opened in the read/write mode, the root
//*	objects are stored in it.  If the root file is read-only,
//*	the data are obtained from it.
//*	The package contains the following classes:
//*	The following parameters can be specified in "jsf.conf". 
//*
//*	    CDCRawData.RFNFormat:       printf format to construct file name
//*	    CDCRawData.DoPlots:         (TRUE/FASLE)
//* (Requires)
//*	TObject:
//*	TClonesArray:
//*	CDCUniversalRecordHeader:
//*	CDCRawADCDatum:
//* (Provides)
//*	CDCBeginRunBuf		: begin run buffer
//*	CDCRawDataBuf		: raw data buffer
//*	CDCEndRunBuf		: end run buffer
//*	CDCRawData		: steering class
//* (Update Record)
//*	1999/06/02  K.Fujii	Original version.
//*************************************************************************

#include "TObject.h"
#include "TClonesArray.h"
#include "TDatime.h"

#include "CDCTypeDef.h"
#include "CDCUniversalRecordHeader.h"
#include "CDCRawADCDatum.h"


// ----------------
// Begin Run Buffer
// ----------------

class CDCRawData;
class CDCBeginRunBuf : public TObject {
protected:
  CDCUniversalRecordHeader  fHeader;	// Header
  Int_t		   	    fDateTime ;	// Date and time
//  CDCHardwareMap	   *fMap;	// (sta#,ch#) --> detector map
//  ADC_t          	   *fThresh;	// threshold data 
  Char_t		   *fComment;	//! Comment string
public:
  CDCBeginRunBuf();
  virtual ~CDCBeginRunBuf();

  Bool_t ReadFile(FILE *fd);
  void   Unpack(BRR_t data[]);
  
  inline CDCUniversalRecordHeader * GetHeader  () { return &fHeader;   }
  inline Int_t			    GetDateTime() const { return  fDateTime; }
  inline Char_t 	     	  * GetComment () const { return  fComment;  }
  void				    DebugPrint ();

  ClassDef(CDCBeginRunBuf, 1) // CDC BeginRun record
};

// ---------------
// Raw Data Buffer
// ---------------

class CDCRawDataBuf : public JSFEventBuf {
private:
  CDCUniversalRecordHeader  fHeader; 	//! Universal Record Header
  Len_t		    	    fNADCData;	//! No. ADC data units
  TClonesArray		   *fADCData;	//! Array of ADC Data
public:
  CDCRawDataBuf(const char *name="CDCRawDataBuf",
		const char *title="CDC raw data", CDCRawData *mod=0);
  CDCRawDataBuf(CDCRawData *mod, const char *name="CDCRawDataBuf",
		const char *title="CDC raw data");
  virtual ~CDCRawDataBuf();

  Bool_t ReadFile(FILE *fd);
  void   Unpack(ADC_t data[]);
  inline CDCUniversalRecordHeader * GetHeader  () { return &fHeader;   }
  inline Len_t			    GetNADCData() const { return  fNADCData; }
  inline TClonesArray		  * GetADCData () const { return  fADCData;  }

  ClassDef(CDCRawDataBuf, 1) // CDC raw data buffer
};

// --------------
// End Run Buffer
// --------------

class CDCEndRunBuf : public TObject {
protected:
  CDCUniversalRecordHeader  fHeader;	// Header 
  Int_t		   	    fDateTime ;	// Date and time
  Char_t		   *fComment;	//! Comment string
public:
  CDCEndRunBuf() {}
  virtual ~CDCEndRunBuf() { if ( fComment ) delete fComment; }

  Bool_t ReadFile(FILE *fd);
  void   Unpack(ERR_t data[]);
  
  inline CDCUniversalRecordHeader * GetHeader  () { return &fHeader;   }
  inline Int_t 	     	     	    GetDateTime() const { return  fDateTime; }
  inline Char_t 	     	  * GetComment () const { return  fComment;  }
  void 			     	    DebugPrint ();

  ClassDef(CDCEndRunBuf, 1) // CDC EndRun record
};

// -----------------
// Raw Data Analysis
// -----------------

class CDCRawData : public JSFModule {
private:
   CDCBeginRunBuf    *fBeginBuf;	//  Begin Run Record
   Char_t   	     *fRFNFormat;	//! Input file name format
   FILE     	     *fd;		//! File descriptor for the raw data

public:
   CDCRawData(const Char_t *name="CDCRawData",
	      const Char_t *title="Get CDC Raw data");
   virtual ~CDCRawData();

   Bool_t BeginRun(Int_t nrun);
   Bool_t Process (Int_t ev);
   void   DebugPrint();

   inline CDCBeginRunBuf    * GetBeginRunBuf() const { return fBeginBuf; }

   ClassDef(CDCRawData, 1) // Read CDC raw data analysis
};

#endif


